"""
Provides support for handling modes that arise from
"""

import enum
import numpy as np, scipy
import McUtils.Numputils as nput
from collections import namedtuple
from McUtils.Coordinerds import CoordinateSystem, CartesianCoordinateSystem3D, InternalCoordinateSystem

# from .MoleculeInterface import AbstractMolecule
# from .Transformations import MolecularTransformation

__all__ = [
    "MixtureModes"
]
class MixtureModes(CoordinateSystem):
    """
    A `McUtils.Coordinerds.CoordinateSystem` object that expresses coordinates as
    a rotation on some base set of coordinates with some associated frequencies.
    """
    name="MixtureModes"
    def __init__(self,
                 basis,
                 coeffs,
                 freqs=None,
                 origin=None,
                 masses=None,
                 inverse=None,
                 mass_weighted=False,
                 frequency_scaled=False,
                 g_matrix=None,
                 name=None,
                 ):
        if (
                isinstance(coeffs, np.ndarray) or
                (not nput.is_numeric(coeffs[0]) and nput.is_numeric(coeffs[0][0]))
        ):
            coeffs = [coeffs]
        coeffs = [np.asanyarray(c) for c in coeffs]
        full_coeffs = coeffs
        coeffs = coeffs[0]
        if inverse is not None:
            inverse = np.asanyarray(inverse)
        if masses is not None:
            masses = np.asanyarray(masses)
        if freqs is not None:
            freqs = np.asanyarray(freqs)
        if g_matrix is not None:
            g_matrix = np.asanyarray(g_matrix)

        super().__init__(
            matrix=coeffs,
            inverse=inverse,
            name=self.name if name is None else name,
            basis=basis,
            dimension=(coeffs.shape[1],),
            origin=origin
        )
        self.freqs = freqs
        self.masses = masses
        self.mass_weighted = mass_weighted
        self.frequency_scaled = frequency_scaled
        self.g_matrix = g_matrix
        self._extended_coeffs = full_coeffs
        self._inverse_coeffs = None


    def to_state(self, serializer=None):
        return {
            'basis':serializer.serialize(self.basis),
            'matrix':self.matrix,
            'inverse':self.inverse,
            'freqs':self.freqs,
            'masses':self.masses,
            'mass_weighted':self.mass_weighted,
            'frequency_scaled':self.frequency_scaled,
            'g_matrix':self.g_matrix
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(
            serializer.deserialize(data['basis']),
            data['matrix'],
            inverse=data['inverse'],
            freqs=data['freqs'],
            masses=data['masses'],
            mass_weighted=data['mass_weighted'],
            frequency_scaled=data['frequency_scaled'],
            g_matrix=data['g_matrix']
        )

    @classmethod
    def prep_modes(cls, modes):
        if isinstance(modes, cls) or all(hasattr(modes, b) for b in ['basis', 'matrix', 'inverse', 'freqs']):
            return modes

        matrix = None
        inverse = None
        basis = None
        freqs = None
        if isinstance(modes, dict):
            opts = modes.copy()
            for k in ["matrix", "modes"]:
                if k in opts:
                    matrix = opts[k]
                    del opts[k]
        elif hasattr(modes, 'matrix'):
            matrix = modes.matrix
            opts = {}

            for k in ['inverse', 'basis', 'freqs']:
                if hasattr(modes, k): opts[k] = getattr(modes, k)
        else:
            matrix = np.asanyarray(modes)
            opts = {}

        if 'inverse' in opts:
            inverse = opts['inverse']
            del opts['inverse']
        if 'basis' in opts:
            basis = opts['basis']
            del opts['basis']
        if 'freqs' in opts:
            freqs = opts['freqs']
            del opts['freqs']

        if matrix is None and inverse is None:
            raise ValueError(f"can't prep {cls.__name__} without matrix or inverse")

        if basis is None:
            if matrix.shape[0] % 3 == 0:
                basis = CartesianCoordinateSystem3D
            else:
                basis = InternalCoordinateSystem(dimension=(None, matrix.shape[0]))

        return cls(
            basis,
            matrix,
            inverse=inverse,
            freqs=freqs,
            **opts
        )

    def __getitem__(self, item):
        """
        Takes a slice of the modes
        :param item:
        :type item:
        :return:
        :rtype:
        """

        if isinstance(item, (int, np.integer)):
            item = (item,)
        elif isinstance(item, slice):
            ...
            # item = np.arange(self.matrix.shape[1])[item]
        elif not isinstance(item[0], (int, np.integer)):
            item = tuple(item[0])

        sub_modes = self.matrix[:, item]
        inv = self._inv
        if inv is not None:
            inv = inv[item, :]
        freq = self.freqs[item,]
        return self.modify(
            matrix=sub_modes,
            freqs=freq,
            inverse=inv
        )

    def modify(self,
               matrix=None,
               *,
               freqs=None,
               origin=None,
               masses=None,
               inverse=None,
               name=None,
               mass_weighted=None,
               frequency_scaled=None,
               g_matrix=None
               ):
        return type(self)(
            self.basis,
            self.matrix if matrix is None else matrix,
            freqs=self.freqs if freqs is None else freqs,
            origin=self.origin if origin is None else origin,
            masses=self.masses if masses is None else masses,
            inverse=self.inverse if inverse is None else inverse,
            name=self.name if name is None else name,
            mass_weighted=self.mass_weighted if mass_weighted is None else mass_weighted,
            frequency_scaled=self.frequency_scaled if frequency_scaled is None else frequency_scaled,
            g_matrix=self.g_matrix if g_matrix is None else g_matrix,
        )

    def rotate(self, rot, in_place=False):
        raise NotImplementedError("too confusing...")

    def transform(self, tf, inv=None, origin=None):
        raise NotImplementedError("ambiguous...")
        #TODO: handle Cartesian variant where tf gets broadcasted

        if inv is None:
            if tf.shape[0] != tf.shape[1]:
                raise ValueError("if given only a transformation, need corresponding inverse")
            else:
                inv = np.linalg.inv(tf)
        base_inv = self.inverse
        base_mat = self.matrix
        raise Exception(base_inv @ base_mat, base_mat.shape, base_inv.shape)
        new_mat = tf@base_mat
        new_inv = base_inv@inv

        if origin is None:
            origin = self.origin
            if origin is not None:
                origin = tf@origin.flatten()

        return type(self)(
            self.basis,
            new_mat,
            freqs=self.freqs,
            masses=self.masses,
            origin=origin,
            inverse=new_inv
        )

    @property
    def cartesian_modes(self):
        return self.origin.ndim == 2

    def embed_coords(self, carts):
        flat_carts = (carts - self.origin[np.newaxis]).reshape((len(carts), -1))
        return (flat_carts[:, np.newaxis, :] @ self.inverse.T[np.newaxis]).reshape(
            flat_carts.shape[0],
            self.matrix.shape[1]
        )
    def unembed_coords(self, mode_coords):
        origin = self.origin
        carts = (mode_coords[:, np.newaxis, :] @ self.matrix.T[np.newaxis]).reshape(
            mode_coords.shape[:1] + origin.shape
        )
        carts = carts + origin[np.newaxis]
        return carts

    @property
    def total_transformation(self):
        return self._extended_coeffs
    @property
    def inverse_transformation(self):
        if self._inverse_coeffs is None:
            self._inverse_coeffs = nput.inverse_transformation(self.total_transformation,
                                                               len(self.total_transformation),
                                                               reverse_expansion=[self.inverse])
        return self._inverse_coeffs
    def embed_derivs(self, derivs):
        return nput.tensor_reexpand(self.total_transformation, derivs)
    def unembed_derivs(self, derivs):
        return nput.tensor_reexpand(self.inverse_transformation, derivs)

    @property
    def is_cartesian(self):
        if self.masses is not None:
            return self.matrix.shape[0] // 3 == len(self.masses)
        else:
            return 'Cartesian' in self.basis.name
    @property
    def coords_by_modes(self):
        return self.inverse
    @property
    def modes_by_coords(self):
        return self.matrix

    def _eval_G(self, masses):
        if not self.is_cartesian:  # a hack
            if self.mass_weighted:
                G = self.inverse.T @ self.inverse
            else:
                # if self.origin is None:
                #     raise ValueError("can't get mass weighting matrix (G^-1/2) without structure")
                # G = self.modes_by_coords @
                raise NotImplementedError("non-mass-weighted internal G-matrix not supported")
        else:
            G = np.diag(1 / np.repeat(masses, 3))
        return G

    def _get_gmatrix(self, masses=None):
        if masses is None:
            G = self.g_matrix
        else:
            G = None
        if G is None:
            if masses is None:
                masses = self.masses
                G = self.g_matrix = self._eval_G(self.masses)
            else:
                G = self._eval_G(self.masses)
        g12 = nput.fractional_power(G, 1/2)
        gi12 = nput.fractional_power(G, -1/2)
        return masses, g12, gi12

    @classmethod
    def compute_local_transformations(cls, f, g):
        return [
            np.diag(np.power(np.diag(g) / np.diag(f), 1 / 4)),
            np.diag(np.power(np.diag(f) / np.diag(g), 1 / 4))
        ]

    @classmethod
    def compute_local_hessian(cls, f, g):
        a = np.diag(np.power(np.diag(g) / np.diag(f), 1 / 4))
        return a @ f @ a

    @classmethod
    def compute_local_gmatrix(cls, f, g):
        a = np.diag(np.power(np.diag(f) / np.diag(g), 1 / 4))
        return a @ g @ a

    def compute_hessian(self, system='modes'):
        if system == 'modes':
            pinv = self.coords_by_modes @ self.modes_by_coords
            return pinv @ np.diag(np.sign(self.freqs) * (self.freqs ** 2)) @ pinv.T
        elif system == 'coords':
            pinv = self.modes_by_coords
            return pinv @ np.diag(np.sign(self.freqs) * (self.freqs ** 2)) @ pinv.T
        else:
            raise ValueError(f'unknown system for normal modes "{system}", valid are "modes", "coords"')

    def compute_gmatrix(self, system='modes', return_fractional=False):
        if system == 'modes':
            if self.mass_weighted:
                g = self.modes_by_coords.T @ self.modes_by_coords
            elif self.g_matrix is not None:
                g = self.modes_by_coords.T @ self.g_matrix @ self.modes_by_coords
            elif self.is_cartesian:
                g = self.modes_by_coords.T @ np.diag(np.repeat(1/self.masses, 3))  @ self.modes_by_coords
            else:
                return None
            if return_fractional:
                g12 = nput.fractional_power(g, 1/2)
                gi12 = nput.fractional_power(g, -1/2)
                return g, g12, gi12
            else:
                return g
                # raise NotImplementedError("non-mass-weighted internal G-matrix not supported")
        elif system == 'coords':
            g, g12, gi12 = self._get_gmatrix()
            if return_fractional:
                return g, g12, gi12
            else:
                return g
        else:
            raise ValueError(f'unknown system for normal modes "{system}", valid are "modes", "coords"')
    def compute_freqs(self):
        # self = self.remove_frequency_scaling().remove_mass_weighting()
        f = self.compute_hessian()
        g = self.compute_gmatrix()
        freqs2, _ = scipy.linalg.eigh(f, g, type=3)
        return np.sign(freqs2) * np.sqrt(np.abs(freqs2))
    @property
    def local_hessian(self):
        # self = self.remove_frequency_scaling()
        f = self.compute_hessian()
        g = self.compute_gmatrix()
        return self.compute_local_hessian(f, g)

    @property
    def local_gmatrix(self):
        # self = self.remove_frequency_scaling()
        f = self.compute_hessian()
        g = self.compute_gmatrix()
        return self.compute_local_gmatrix(f, g)

    @property
    def local_freqs(self):
        return np.diag(self.local_hessian)

    @property
    def local_mode_transformations(self):
        f = self.compute_hessian()
        g = self.compute_gmatrix()
        return self.compute_local_transformations(f, g)


    def get_nearest_mode_transform(self,
                                   alternate_modes:np.ndarray,
                                   mass_weighted=None,
                                   atoms=None,
                                   maximum_similarity=True,
                                   unitarize=True,
                                   masses=None
                                   ):
        if mass_weighted is not None:
            if not mass_weighted:
                if masses is None:
                    masses = self.masses
                    modes = self.remove_mass_weighting()
                else:
                    modes = self.remove_mass_weighting(masses)
            else:
                if masses is None:
                    masses = self.masses
                    modes = self.make_mass_weighted()
                else:
                    modes = self.make_mass_weighted(masses)
        else:
            modes = self

        inv = modes.coords_by_modes
        modes = modes.modes_by_coords
        if self.is_cartesian and atoms is not None:
            proj = self._atom_projector(modes.shape[0], atoms)
            # if mass_weighted:
            #     gi12 = np.diag(np.repeat(1/np.sqrt(masses), 3))
            #     proj = nput.find_basis(gi12 @ proj @ gi12)
            modes = proj @ modes
            alternate_modes = proj @ alternate_modes

        if maximum_similarity:
            tf = nput.maximum_similarity_transformation(modes, alternate_modes, apply_transformation=False)
            return tf, tf.T
        else:
            tf = inv @ alternate_modes
            if unitarize:
                tf = nput.unitarize_transformation(tf)
                inv = tf.T
            else:
                inv = np.linalg.pinv(tf)
            return tf, inv

    localization_type = 'ned'
    zero_freq_cutoff = 99 / 219474.56
    def get_projected_localized_mode_transformation(self,
                                                    projectors,
                                                    masses=None, origin=None,
                                                    localization_type=None,
                                                    allow_mode_mixing=False,
                                                    maximum_similarity=False,
                                                    unitarize=True,
                                                    zero_freq_cutoff=None,
                                                    orthogonal_projection=False,
                                                    atoms=None #ignored but tedious to have around
                                                    ):
        if zero_freq_cutoff is None:
            zero_freq_cutoff = self.zero_freq_cutoff
        if localization_type is None:
            localization_type = self.localization_type

        if masses is None:
            masses = self.masses
            mw = self.make_mass_weighted()
        else:
            mw = self.make_mass_weighted(masses=masses)

        f = mw.compute_hessian('coords')

        if self.is_cartesian and localization_type == 'direct':
            if origin is None:
                origin = mw.remove_mass_weighting().origin
            tr_tf, tr_inv = nput.translation_rotation_invariant_transformation(
                origin.reshape(-1, 3),
                masses,
                mass_weighted=True
            )
        else:
            tr_tf = tr_inv = np.eye(mw.matrix.shape[0])

        projectors = [
            (
                nput.orthogonal_projection_matrix(p)
                    if orthogonal_projection else
                nput.projection_matrix(p)
            )
                if p.shape[0] != p.shape[1] else
            np.asanyarray(p)
            for p in projectors
        ]

        g_matrix = tr_tf.T @ tr_tf
        if allow_mode_mixing:
            proj = np.sum(projectors, axis=0)
            # proj = proj @ g12
            # proj = self._atom_projector(len(masses), atoms)
            freqs, modes, inv = self.get_normal_modes(
                tr_inv @ proj @ f @ proj.T @ tr_inv.T,
                g_matrix,
                remove_transrot=True,
                mass_weighted=True,
                zero_freq_cutoff=zero_freq_cutoff
            )
            # freqs1, _, _ = self.get_normal_modes(
            #      f,
            #     g_matrix,
            #     remove_transrot=True,
            #     mass_weighted=True,
            #     zero_freq_cutoff=zero_freq_cutoff
            # )
            # raise Exception(freqs * 219474.63, freqs1 * 219474.63)
        else:
            # freqs = []
            modes = []
            # inv = []
            for proj in projectors:
                f_matrix = tr_inv @ proj @ f @ proj.T @ tr_inv.T
                # import McUtils.Formatters as mfmt
                # print(
                #     mfmt.TableFormatter("{:.3f}").format(
                #         f_matrix * 219474.56
                #     )
                # )
                sub_freqs, sub_modes, sub_inv = self.get_normal_modes(
                    f_matrix,
                    g_matrix,
                    remove_transrot=True,
                    mass_weighted=True,
                    zero_freq_cutoff=zero_freq_cutoff
                )
                # print(sub_freqs * 219474.56)

                # freqs.append(sub_freqs)
                modes.append(sub_modes)
                # inv.append(sub_inv)
            # freqs = np.concatenate(freqs)
            modes = np.concatenate(modes, axis=1)
            # inv = np.concatenate(inv, axis=0)

        if localization_type == 'direct':
            modes = tr_tf @ modes

        return mw.get_nearest_mode_transform(
            modes,
            mass_weighted=True,
            maximum_similarity=maximum_similarity,
            unitarize=unitarize
        )

    def get_atom_localized_mode_transformation(self,
                                               atoms,
                                               masses=None, origin=None,
                                               localization_type='ned',
                                               allow_mode_mixing=False,
                                               maximum_similarity=False,
                                               orthogonal_projection=False,
                                               unitarize=True,
                                               zero_freq_cutoff=None
                                               ):
        if nput.is_numeric(atoms):
            atoms = [atoms]
        if masses is None:
            m = self.masses
        else:
            m = masses
        nats = len(m)
        if orthogonal_projection:
            atoms = np.setdiff1d(np.arange(nats), atoms)

        proj = (
            [self._atom_projector(nats, atoms)]
                if allow_mode_mixing else
            [self._atom_projector(nats, a) for a in atoms]
        )
        # gi12 = np.diag(np.repeat(1 / np.sqrt(m), 3))
        # proj = nput.find_basis(gi12 @ proj @ gi12)

        return self.get_projected_localized_mode_transformation(
            proj,
            origin=origin,
            masses=masses,
            localization_type=localization_type,
            maximum_similarity=maximum_similarity,
            unitarize=unitarize,
            allow_mode_mixing=allow_mode_mixing,
            zero_freq_cutoff=zero_freq_cutoff
        )

    def get_coordinate_projected_localized_mode_transformation(self,
                                                               coordinate_constraints,
                                                               atoms=None,
                                                               masses=None, origin=None,
                                                               localization_type='ned',
                                                               allow_mode_mixing=False,
                                                               maximum_similarity=False,
                                                               orthogonal_projection=True,
                                                               unitarize=True
                                                               ):


        if self.is_cartesian:
            if nput.is_numeric(coordinate_constraints[0]):
                coordinate_constraints = [coordinate_constraints]
            nmw = self.remove_mass_weighting()
            if masses is None:
                m = self.masses
            else:
                m = masses
            basis, _, _ = nput.internal_basis(nmw.origin.reshape((-1, 3)), coordinate_constraints, masses=m)

            g12 = np.diag(np.repeat(1/np.sqrt(m), 3))
            projections = [
                nput.projection_matrix(g12 @ b)
                    if not orthogonal_projection else
                nput.orthogonal_projection_matrix(g12 @ b)
                for b in basis
            ]

            if allow_mode_mixing and orthogonal_projection:
                proj = projections[0]
                for p in projections[1:]:
                    proj = p @ proj @ p
                projections = [proj]

            if atoms is not None:
                if nput.is_numeric(atoms):
                    atoms = [atoms]
                nats = len(m)
                a_proj = self._atom_projector(nats, atoms)
                projections = [
                    a_proj @ proj @ a_proj
                    for proj in projections
                ]
        else:
            if nput.is_numeric(coordinate_constraints):
                coordinate_constraints = [coordinate_constraints]
            if not (
                # Fast check
                isinstance(coordinate_constraints, np.ndarray)
                and np.issubdtype(coordinate_constraints.dtype, np.integer)
            ) and not all(nput.is_int(c) for c in coordinate_constraints):
                raise ValueError(f"internal modes can only be constrained by coordinate index (got {coordinate_constraints})")
            basis = np.zeros((len(coordinate_constraints),self.coords_by_modes.shape[1] ), dtype=float)
            coordinate_constraints = np.asanyarray(coordinate_constraints)
            for n,i in enumerate(coordinate_constraints):
                basis[n, i] = 1


            # g12 = nput.fractional_power(self.g_matrix, 1/2)
            # projections = [
            #     nput.projection_matrix(g12 @ b[:, np.newaxis])
            #         if not orthogonal_projection else
            #     nput.orthogonal_projection_matrix(g12 @ b[:, np.newaxis])
            #     for b in basis
            # ]

            projections = [
                nput.projection_matrix(b[:, np.newaxis])
                    if not orthogonal_projection else
                nput.orthogonal_projection_matrix(b[:, np.newaxis])
                for b in basis
            ]

        # if orthogonal_projection:
        #     projections = [proj @ gi12 for proj in projections]
        # else:
        #     projections = [proj @ gi12 for proj in projections]


        return self.get_projected_localized_mode_transformation(
            projections,
            origin=origin,
            masses=masses,
            localization_type=localization_type,
            maximum_similarity=maximum_similarity,
            unitarize=unitarize,
            allow_mode_mixing=allow_mode_mixing
        )

    def get_internal_localized_mode_transformation(
            self,
            expansion_coordinates: "Iterable[Iterable[int]|dict]",
            fixed_atoms=None,
            mass_weighted=False,
            project_transrot=True,
            atoms=None,
            maximum_similarity=False,
            orthogonal_projection=False,
            projection=False,
            allow_mode_mixing=False,
            unitarize=True,
            origin=None,
            masses=None,
            localization_type='ned'
    ):
        # from .ObliqueModes import ObliqueModeGenerator

        nmw = self.remove_mass_weighting()
        base_derivs = nput.internal_coordinate_tensors(
            nmw.origin,
            expansion_coordinates,
            fixed_atoms=fixed_atoms
        )

        if projection or orthogonal_projection:
            if masses is None:
                m = self.masses
            else:
                m = masses
            base_derivs = np.diag(np.repeat(1 / np.sqrt(m), 3)) @ base_derivs

            if origin is None:
                origin = self.origin
            if masses is None:
                m = self.masses
            else:
                m = masses
            projector = nput.translation_rotation_projector(
                origin.reshape(-1, 3),
                m,
                mass_weighted=True
            )
            base_derivs = projector @ base_derivs

            # g12 = np.diag(np.repeat(np.sqrt(m), 3))
            gi12 = np.diag(np.repeat(1 / np.sqrt(m), 3))
            if allow_mode_mixing:
                projections = [
                    nput.projection_matrix(gi12 @ base_derivs)
                        if not orthogonal_projection else
                    nput.orthogonal_projection_matrix(gi12 @ base_derivs)
                ]
            else:
                projections = [
                    nput.projection_matrix(gi12 @ b[:, np.newaxis])
                        if not orthogonal_projection else
                    nput.orthogonal_projection_matrix(gi12 @ b[:, np.newaxis])
                    for b in base_derivs.T
                ]

            # if orthogonal_projection:
            #     projections = [proj @ gi12 for proj in projections]
            # else:
            #     projections = [proj @ gi12 for proj in projections]

            if atoms is not None:
                if nput.is_numeric(atoms):
                    atoms = [atoms]
                nats = len(m)
                a_proj = self._atom_projector(nats, atoms)
                projections = [
                    a_proj @ proj @ a_proj
                    for proj in projections
                ]

            return self.get_projected_localized_mode_transformation(
                projections,
                origin=origin,
                masses=masses,
                localization_type=localization_type,
                maximum_similarity=maximum_similarity,
                unitarize=unitarize,
                allow_mode_mixing=allow_mode_mixing
            )

        else:

            if mass_weighted:
                if masses is None:
                    m = self.masses
                else:
                    m = masses
                base_derivs = np.diag(np.repeat(1 / np.sqrt(m), 3)) @ base_derivs

            if project_transrot:
                if origin is None:
                    origin = self.origin
                if masses is None:
                    m = self.masses
                else:
                    m = masses
                projector = nput.translation_rotation_projector(
                        origin.reshape(-1, 3),
                        m,
                        mass_weighted=mass_weighted
                    )
                base_derivs = projector @ base_derivs

            return self.get_nearest_mode_transform(
                base_derivs,
                atoms=atoms,
                mass_weighted=mass_weighted,
                maximum_similarity=maximum_similarity,
                unitarize=unitarize
            )

    def get_displacement_localized_mode_transformation(self,
                                                       mode_blocks=None,
                                                       atoms=None,
                                                       mass_weighted=True,
                                                       unitarize=True,
                                                       **maximizer_opts
                                                       ):
        if not unitarize:
            raise ValueError("Jacobi iterations only give a unitary transformation")

        if mass_weighted:
            modes = self.make_mass_weighted().matrix
        else:
            modes = self.remove_mass_weighting().matrix

        if atoms is not None:
            modes = self._atom_projector(modes.shape[0], atoms) @ modes
        if mode_blocks is None:
            mode_blocks = [np.arange(modes.shape[1])]
        elif nput.is_numeric(mode_blocks[0]):
            mode_blocks = [mode_blocks]

        tf_inv = np.zeros((modes.shape[1], modes.shape[1]), dtype=float)
        for block in mode_blocks:
            _, sub_inv, _ = nput.jacobi_maximize(
                modes[:, block],
                nput.displacement_localizing_rotation_generator,
                **maximizer_opts
            )
            tf_inv[np.ix_(block, block)] = sub_inv

        return tf_inv, tf_inv.T

    def get_mass_scaled_mode_transformation(self,
                                            mass_scaling,
                                            *,
                                            atoms,
                                            localization_cutoff=.8,
                                            num_modes=None,
                                            project_transrot=False,
                                            unitarize=True,
                                            **diag_opts
                                            ):
        from .NormalModes import NormalModes
        self = self.make_mass_weighted()

        f = self.compute_hessian('coords')

        if not nput.is_numeric(mass_scaling):
            mass_scaling = np.asanyarray(mass_scaling)

        if nput.is_int(atoms):
            atoms = [atoms]
        atoms = tuple(atoms)

        scaled_masses = np.array(self.masses)
        scaled_masses[atoms,] *= mass_scaling

        if project_transrot:
            proj = nput.translation_rotation_projector(
                self.remove_mass_weighting().origin,
                masses=scaled_masses,
                mass_weighted=True
            )
        else:
            proj = None
        freqs0, q1, _ = NormalModes.get_normal_modes(f, scaled_masses,
                                                     projector=proj,
                                                     mass_weighted=True,
                                                     **diag_opts
                                                     )

        atom_inds = np.concatenate([
            n * 3 + np.arange(3)
            for n in atoms
        ])
        atom_proj = q1[atom_inds, :]
        max_local_modes = np.linalg.norm(atom_proj, axis=0)

        max_num_modes = len(atoms) * 3
        if localization_cutoff is None:
            num_modes = max_num_modes if num_modes is None else num_modes
            mode_pos = np.argsort(-max_local_modes)[:num_modes]
        else:
            mode_pos = np.where(max_local_modes > localization_cutoff)
            if len(mode_pos) == 0 or len(mode_pos[0]) == 0:
                return None, None

            mode_vals = max_local_modes[mode_pos]
            num_modes = max_num_modes if num_modes is None else num_modes
            ord = np.argsort(-mode_vals)[:num_modes]
            mode_pos = mode_pos[0][ord,]

        tf = self.coords_by_modes@q1[:, mode_pos]
        ord2 = np.argsort(np.argmax(np.abs(tf), axis=0))
        tf = tf[:, ord2]

        return tf, tf.T


    class LocalizationMethods(enum.Enum):
        MaximumSimilarity = 'target_modes'
        AtomLocalized = 'atoms'
        DisplacmentMinimized = 'displacements'
        Internals = 'coordinates'
        CoordinateConstraints = 'constraints'
        Projected = 'projections'
        MassScaled = 'mass_scaling'

    @property
    def localizer_dispatch(self):
        return {
            self.LocalizationMethods.MaximumSimilarity.value:(self.get_nearest_mode_transform, 'target_modes'),
            self.LocalizationMethods.AtomLocalized.value:(self.get_atom_localized_mode_transformation, 'atoms'),
            self.LocalizationMethods.DisplacmentMinimized.value:(self.get_displacement_localized_mode_transformation, 'mode_blocks'),
            self.LocalizationMethods.Internals.value:(self.get_internal_localized_mode_transformation, 'internals'),
            self.LocalizationMethods.CoordinateConstraints.value:(self.get_coordinate_projected_localized_mode_transformation, 'coordinate_constraints'),
            self.LocalizationMethods.Projected.value:(self.get_projected_localized_mode_transformation, 'projections'),
            self.LocalizationMethods.MassScaled.value:(self.get_mass_scaled_mode_transformation, 'mass_scaling'),
        }

    def localize(self,
                 method=None,
                 *,
                 atoms=None,
                 target_modes=None,
                 internals=None,
                 mode_blocks=None,
                 coordinate_constraints=None,
                 projections=None,
                 reorthogonalize=None,
                 mass_scaling=None,
                 unitarize=True,
                 **opts
                 ):

        from .LocalizedModes import LocalizedModes

        if method is None:
            if target_modes is not None:
                method = 'target_modes'
            elif internals is not None:
                method = 'coordinates'
            elif mode_blocks is not None:
                method = 'displacements'
            elif mass_scaling is not None:
                method = 'mass_scaling'
            elif atoms is not None:
                method = 'atoms'
            elif coordinate_constraints is not None:
                method = 'constraints'
            elif projections is not None:
                method = 'projections'
            else:
                method = 'displacements'

        if isinstance(method, str):
            method = self.LocalizationMethods(method)

        if hasattr(method, 'value'):
            method = self.localizer_dispatch.get(method.value, method)

        args = ()
        try:
            method, arg_names = method
        except TypeError:
            ...
        else:
            if isinstance(arg_names, str):
                arg_names = [arg_names]
            all_kw = dict(
                atoms=atoms,
                target_modes=target_modes,
                internals=internals,
                mode_blocks=mode_blocks,
                reorthogonalize=reorthogonalize,
                coordinate_constraints=coordinate_constraints,
                unitarize=unitarize,
                projections=projections,
                mass_scaling=mass_scaling
            )
            args = tuple(all_kw.get(k) for k in arg_names)
            if 'atoms' not in arg_names:
                opts['atoms'] = atoms # taken by all of the localizers

        tf, inv = method(*args, unitarize=unitarize, **opts)
        if tf is None: return None
        # inverse = tf.T @ self.inverse # assumes a unitary localization

        if reorthogonalize is None:
            reorthogonalize = not unitarize
        if reorthogonalize:
            modes = self.make_mass_weighted().matrix @ tf
            tf = tf @ nput.fractional_power(modes.T @ modes, -1 / 2)
            inv = nput.fractional_power(modes.T @ modes, 1 / 2) @ inv

        return LocalizedModes(self, tf, inverse=inv)

    def make_mass_weighted(self, masses=None):
        if self.mass_weighted: return self
        masses, g12, gi12 = self._get_gmatrix(masses=masses)
        L = g12 @ self.modes_by_coords
        Linv = self.coords_by_modes @ gi12
        origin = (self.origin.flatten()[np.newaxis, :] @ gi12).reshape(self.origin.shape)

        return self.modify(L,
                           inverse=Linv,
                           masses=masses,
                           origin=origin,
                           mass_weighted=True
                           )
    def remove_mass_weighting(self, masses=None):
        if not self.mass_weighted: return self
        masses, g12, gi12 = self._get_gmatrix(masses=masses)
        L = gi12 @ self.modes_by_coords
        Linv = self.coords_by_modes @ g12
        origin = (self.origin.flatten()[np.newaxis, :] @ g12).reshape(self.origin.shape)

        return self.modify(L,
                           inverse=Linv,
                           masses=masses,
                           origin=origin,
                           mass_weighted=False
                           )

    def _frequency_scaling(self, freqs=None):
        # L = self.matrix.shape.T
        if freqs is None:
            freqs = self.freqs
        conv = np.sign(freqs) * np.sqrt(np.abs(freqs))
        return freqs, conv

    def make_frequency_scaled(self, freqs=None):
        if self.frequency_scaled: return self
        freqs, conv = self._frequency_scaling(freqs=freqs)
        L = self.matrix * conv[np.newaxis, :]
        Linv = self.inverse / conv[:, np.newaxis] # del_Q X

        return self.modify(L,
                           inverse=Linv,
                           freqs=freqs,
                           frequency_scaled=True
                           )

    def remove_frequency_scaling(self, freqs=None):
        if not self.frequency_scaled: return self
        freqs, conv = self._frequency_scaling(freqs=freqs)
        L = self.matrix / conv[np.newaxis, :]
        Linv = self.inverse * conv[:, np.newaxis] # del_Q X

        return self.modify(L,
                           inverse=Linv,
                           freqs=freqs,
                           frequency_scaled=False
                           )

    def make_dimensionless(self, freqs=None, masses=None):
        new = self
        if not new.mass_weighted: new = new.make_mass_weighted(masses=masses)
        if not new.frequency_scaled: new = new.make_frequency_scaled(freqs=freqs)

        return new

    def make_dimensioned(self, freqs=None, masses=None):
        new = self
        if not new.mass_weighted: new = new.remove_mass_weighting(masses=masses)
        if not new.frequency_scaled: new = new.remove_frequency_scaling(freqs=freqs)
        return new

    # def take_submodes(self, pos, axes=(0, 1, 2)):
    #     if self.is_cartesian:
    #         pos = np.asanyarray(pos)
    #         axes = np.asanyarray(axes)
    #         pos = (axes[np.newaxis, :] + pos[:, np.newaxis]*len(axes)).flatten()
    #     atom_disps = self.matrix[pos]
    #     f_base_sub = atom_disps.T @ np.diag(self.freqs**2) @ atom_disps
    #     return type(self).from_fg(
    #         self.basis.take_subbasis(pos),
    #         f_base_sub,
    #         self.g_matrix[np.ix_(pos, pos)],
    #         ...
    #     )

    def apply_projection(self, proj, project_transrot=True, masses=None, origin=None):
        if project_transrot:
            if masses is None:
                m = self.masses
            else:
                m = masses
            if origin is None:
                o = self.origin
            else:
                o = origin

            tr_proj = nput.translation_rotation_projector(
                np.asanyarray(o).reshape((-1, 3)),
                masses=m,
                mass_weighted=self.mass_weighted
            )
            proj = tr_proj @ proj @ tr_proj



        mbc = proj @ self.modes_by_coords
        cbm = self.coords_by_modes @ proj

        return self.modify(
            matrix=mbc,
            inverse=cbm,
            masses=masses,
            origin=origin
        )

    @classmethod
    def _atom_projector(cls, n, i, orthogonal_projection=False):
        if nput.is_numeric(i):
            i = [i]
        z = np.zeros((3 * n, 3 * n))
        for i in i:
            x = np.arange(3 * i, 3 * (i + 1))
            z[x, x] = 1
        if orthogonal_projection:
            z = np.eye(3 * n) - z
        return z
    def apply_constraints(self,
                          coordinate_constraints,
                          atoms=None,
                          masses=None,
                          origin=None,
                          orthogonal_projection=True,
                          ):

        if nput.is_numeric(coordinate_constraints[0]):
            coordinate_constraints = [coordinate_constraints]

        if masses is not None:
            m = masses
        else:
            m = self.masses
        if origin is not None:
            o = origin
        else:
            o = self.remove_mass_weighting().origin
        basis, _, _ = nput.internal_basis(
            np.asanyarray(o).reshape((-1, 3)),
            coordinate_constraints,
            masses=m,
            project_transrot=False
            )

        gi12 = np.diag(np.repeat(1 / np.sqrt(m), 3))
        if self.mass_weighted:
            projections = [
                nput.projection_matrix(gi12 @ b)
                    if not orthogonal_projection else
                nput.orthogonal_projection_matrix(gi12 @ b)
                for b in basis
            ]
        else:
            projections = [
                nput.projection_matrix(b)
                    if not orthogonal_projection else
                nput.orthogonal_projection_matrix(b)
                for b in basis
            ]

        if atoms is not None:
            if nput.is_numeric(atoms):
                atoms = [atoms]
            nats = len(self.masses)
            a_proj = self._atom_projector(nats, atoms)
            projections = [
                a_proj @ proj @ a_proj
                for proj in projections
            ]

        if orthogonal_projection:
            proj = projections[0]
            for p in projections[1:]:
                proj = p @ proj @ p
        else:
            proj = np.sum(projections, axis=0)

        new = self.apply_projection(proj, masses=m, origin=o, project_transrot=False)
        # if self.mass_weighted:
        #     new = new.make_mass_weighted(masses=m)

        return new

    def apply_transformation(self, tf, **opts):
        from .LocalizedModes import LocalizedModes

        return LocalizedModes(self, tf, **opts)

    def make_oblique(self):
        from .ObliqueModes import ObliqueModeGenerator

        f = self.compute_hessian()
        g = self.compute_gmatrix()

        _, _, u, ui = ObliqueModeGenerator(f, g).run()
        return self.apply_transformation((u, ui))





import numpy as np
from .BaseSpectrum import DiscreteSpectrum

__all__ = [
    "HarmonicSpectrum"
]

class HarmonicSpectrum(DiscreteSpectrum):

    @classmethod
    def from_normal_modes(cls, nms, dipole_derivatives, **opts):
        nms = nms.remove_mass_weighting().make_frequency_scaled()
        freqs = nms.freqs
        tms = 1/np.sqrt(2) * nms.coords_by_modes @ dipole_derivatives
        return cls.from_transition_moments(
            freqs,
            tms,
            **opts
        )

    @classmethod
    def from_mol(cls, mol, **opts):
        return cls.from_normal_modes(
            mol.get_normal_modes(),
            mol.get_cartesian_dipole_derivatives(1)[0],
            **opts
        )


import numpy as np
import McUtils.Devutils as dev
import McUtils.Numputils as nput
from McUtils.ExternalPrograms import PILInterface
from McUtils.Plots import ColorPalette

__all__ = [
    "SpectrumExtractor"
]

class SpectrumExtractor:

    def __init__(self, image_data, color_space='rgb', operation_color_space='lab'):
        image_data = np.asanyarray(image_data)
        if color_space != operation_color_space:
            image_data = ColorPalette.color_convert(image_data, color_space, operation_color_space)
        self.img = image_data
        self.color_space = operation_color_space

    @classmethod
    def from_pil(cls, pil_image, color_space=None, **etc):
        if color_space is None:
            color_space = pil_image.mode.lower()
            if color_space == 'l':
                color_space = 'grayscale'
            elif color_space == 'rgba':
                color_space = 'rgb'

        image_data = np.moveaxis(np.asarray(pil_image), -1, 0)[:3]
        if color_space == 'rgb':
            image_data = np.clip(image_data, 0, 255)

        return cls(
            image_data,
            color_space=color_space,
            **etc
        )
    @classmethod
    def from_file(cls, file, color_space=None, **etc):
        pil = PILInterface.from_file(file)

        return cls.from_pil(pil, color_space=color_space, **etc)
    @classmethod
    def from_url(cls, file, color_space=None):
        pil = PILInterface.from_url(file)
        return cls.from_pil(pil, color_space=color_space)

    default_tolerances = [25, 25, 25]
    def find_pixels(
            self,
            color,
            distance_tol=None,
            tolerances=None,
            color_space='rgb',
            search_color_space=None,
            image_range=None
    ):
        if isinstance(color, str):
            color = ColorPalette.parse_color_string(color)

        if search_color_space is None:
            search_color_space = self.color_space
        color = ColorPalette.color_convert(color, color_space, search_color_space)

        image_data = self.img
        if search_color_space is not None and search_color_space != self.color_space:
            image_data = ColorPalette.color_convert(image_data, self.color_space, search_color_space)

        if image_range is not None:
            y,x = image_range
            if nput.is_numeric(x):
                y,x = None, [y,x]
            if x is not None:
                xm, xM = x
                if not nput.is_int(xm):
                    xm = int(image_data.shape[-2] * xm)
                if not nput.is_int(xM):
                    xM = int(image_data.shape[-2] * xM)
                image_data = image_data[:, xm:xM, :]
            if y is not None:
                ym, yM = y
                if not nput.is_int(ym):
                    ym = int(image_data.shape[-1] * ym)
                if not nput.is_int(yM):
                    yM = int(image_data.shape[-1] * yM)
                image_data = image_data[:, :, ym:yM]

        if distance_tol is not None:
            raise NotImplementedError("finding pixels from color distance not supported")
        else:
            if tolerances is None:
                tolerances = self.default_tolerances

            mask = np.full(image_data[0].shape, True)
            for channel, center, tol in zip(
                image_data,
                color,
                tolerances
            ):
                mask = np.logical_and(
                    mask,
                    np.logical_and(
                        channel >= center - tol,
                        channel <= center + tol
                    )
                )

        return np.where(mask)

    def find_spectrum_lines(self,
                            pixel_positions,
                            max_pixel_distance=.005,
                            min_line_cutoff=0.5,
                            smoothing=True,
                            line_split_cutoff=5,
                            allow_disjoint=False,
                            spectrum_direction='up'
                            ):

        if max_pixel_distance < 1:
            max_pixel_distance = max_pixel_distance * max(self.img.shape[-1:])

        lines = []
        for x,Y in zip(*nput.group_by(pixel_positions[1], pixel_positions[0])[0]):
            Y = np.sort(Y) # just in case
            # if smoothing:
            y_diffs = np.diff(Y)
            split_points = np.where(np.abs(y_diffs) > line_split_cutoff)
            if len(split_points) > 0 and len(split_points[0]) > 0:
                Y = [
                    g
                    for g in np.split(Y, split_points[0]+1)
                ]
            else:
                Y = [Y]

            if spectrum_direction.lower() == 'down':
                Y = reversed(Y)

            if len(lines) == 0:
                lines.extend([[x, y]] for y in Y)
            else:
                add_pos = set()
                for y in Y:
                    for n,l in enumerate(lines):
                        if n in add_pos: continue

                        xx, yy = l[-1]
                        dx = x-xx
                        if abs(dx) > max_pixel_distance: continue

                        dy = min([abs(y[0] - yy[0]), abs(y[-1] - yy[-1])]) # check boundaries first
                        if abs(dx)+abs(dy) > max_pixel_distance:
                            dy = np.min(np.abs(yy[:, np.newaxis] - y[np.newaxis, :]))
                        if abs(dx)+abs(dy) < max_pixel_distance:
                            l.append([x, y])
                            if not allow_disjoint:
                                add_pos.add(n)
                            break
                    else:
                        lines.append([[x, y]])


        if min_line_cutoff < 1:
            min_line_cutoff = self.img.shape[-1] * min_line_cutoff
            lines = [
                ll for ll in lines
                if len(ll) > min_line_cutoff
            ]


        if smoothing:
            lines = [
                [[x, np.average(y)] for x, y in l]
                for l in lines
            ]
            if nput.is_int(smoothing) and smoothing is not True:
                lines = [np.array(l).T for l in lines]
                lines = [
                    np.array([
                        np.convolve(x, np.ones(smoothing), mode='valid') / smoothing,
                        np.convolve(y, np.ones(smoothing), mode='valid') / smoothing
                    ]).T
                    for x,y in lines
                ]
        else:
            lines = [
                sum(
                    ([[x, yy] for yy in y] for x, y in l),
                    []
                )
                for l in lines
            ]

        return [
            np.array(l).T
            for l in sorted(lines, key=lambda x:-len(x))
        ]

    default_merge_tolerances = [10, 10, 10]
    def get_dominant_colors(self, bins=255, color_space='lab', min_counts=2, merge_tolerances=None):
        image_data = self.img
        if color_space != self.color_space:
            image_data = ColorPalette.color_convert(image_data, self.color_space, color_space)
        dists, (x, y, z) = np.histogramdd(np.moveaxis(image_data, 0, -1).reshape(-1, 3), bins=bins)
        take = np.where(dists > min_counts)
        base_colors = {
            (
                (x[i] + x[i+1])/2,
                (y[j] + y[j+1])/2,
                (z[k] + z[k+1])/2,
            ):int(dists[i,j,k])
            for i,j,k in zip(*take)
        }

        base_colors = {
            k:base_colors[k]
            for k in sorted(base_colors, key=lambda x:-base_colors[x])
        }
        if merge_tolerances is None:
            merge_tolerances = self.default_merge_tolerances

        test_keys = list(base_colors.keys())
        merges = {}
        for n,c in enumerate(base_colors.keys()):
            for k in test_keys[:max(n-1, 0)]:
                if c == k or k in merges: continue
                if all(abs(k[i] - c[i]) <= merge_tolerances[i] for i in range(3)):
                    merges[c] = k
                    break

        for c,k in merges.items():
            base_colors[k] += base_colors[c]
            del base_colors[c]

        # print(
        #     sum(v for v in base_colors.values()) / (image_data.shape[-1] * image_data.shape[-2])
        # )

        return {
            k:base_colors[k]
            for k in sorted(base_colors, key=lambda x:-base_colors[x])
        }


        #TOOD: move this to OpenCV
        # all_pixels = np.moveaxis(self.img, 0, -1).reshape(-1, 3)
        # pix = {}
        # if tolerances is None:
        #     tolerances = self.default_tolerances
        # tolerances = np.asanyarray(tolerances)
        # m = 0
        # for p in all_pixels:
        #     old_pix = None
        #     new_pix = {}
        #     m += 1
        #     for k in pix:
        #         if all(k[i] - p[i] <= tolerances[i] for i in range(3)):
        #             n = pix[k]
        #             p = tuple((n*k[i] + p[i]) / (n+1) for i in range(3))
        #             new_pix[p] = n + 1
        #             old_pix = k
        #             break
        #     else:
        #         p = tuple(p)
        #         new_pix[p] = 1
        #     if old_pix is not None:
        #         del pix[old_pix]
        #     pix.update(new_pix)
        #
        # return pix

    def identify_frame_x_boundaries(self, pixel_positions, min_line_cutoff=.5, frame_gap_cutoff=.05):
        min_x = None
        cur_x = None
        max_X = None
        if min_line_cutoff < 1:
            min_line_cutoff = self.img.shape[-1] * min_line_cutoff
        if frame_gap_cutoff < 1:
            frame_gap_cutoff = self.img.shape[-1] * frame_gap_cutoff

        miss_counts = 0
        for x,Y in zip(*nput.group_by(pixel_positions[1], pixel_positions[0])[0]):
            if len(Y) > min_line_cutoff:
                if min_x is not None:
                    max_X = x - 1
                    break
                cur_x = x
            elif cur_x is not None:
                miss_counts += 1
                if miss_counts > frame_gap_cutoff:
                    min_x = cur_x + 1

        if min_x is None:
            if max_X is not None:
                if max_X < .5 * self.img.shape[-1]:
                    min_x = max_X
                    max_X = None
            else:
                min_x = np.min(pixel_positions[0])
        if max_X is None:
            max_X = np.max(pixel_positions[0])

        return min_x, max_X

    def identify_frame_boundaries(self, pixel_positions,
                                  min_line_cutoffs=.5,
                                  frame_gap_cutoffs=.05,
                                  identified_components=True
                                  ):
        if nput.is_numeric(min_line_cutoffs): min_line_cutoffs = [min_line_cutoffs, min_line_cutoffs]
        if nput.is_numeric(frame_gap_cutoffs): frame_gap_cutoffs = [frame_gap_cutoffs, frame_gap_cutoffs]
        x_cutoff, y_cutoff = min_line_cutoffs
        x_gap, y_gap = frame_gap_cutoffs

        if x_cutoff < 1:
            x_cutoff = self.img.shape[-2] * x_cutoff
        if y_cutoff < 1:
            y_cutoff = self.img.shape[-1] * y_cutoff

        if x_gap < 1:
            x_gap = self.img.shape[-2] * x_gap
        if y_gap < 1:
            y_gap = self.img.shape[-1] * y_gap

        if identified_components is True: identified_components = [identified_components, identified_components]

        find_x, find_y = identified_components
        if find_x:
            x_range = self.identify_frame_x_boundaries(
                (pixel_positions[0], pixel_positions[1]),
                min_line_cutoff=x_cutoff,
                frame_gap_cutoff=x_gap
            )
        else:
            x_range = [np.min(pixel_positions[0]), np.max(pixel_positions[0])]

        if find_y:
            y_range = self.identify_frame_x_boundaries(
                (pixel_positions[1], pixel_positions[0]),
                min_line_cutoff=y_cutoff,
                frame_gap_cutoff=y_gap
            )
        else:
            y_range = [np.min(pixel_positions[1]), np.max(pixel_positions[1])]

        return x_range, y_range

    def prune_straight_vertical_pixels(self, pixel_positions, min_line_cutoff=.5):
        new_x = []
        new_y = []
        if min_line_cutoff < 1:
            min_line_cutoff = self.img.shape[-1] * min_line_cutoff
        for x,Y in zip(*nput.group_by(pixel_positions[1], pixel_positions[0])[0]):
            if len(Y) > min_line_cutoff: continue
            new_x.append(np.full(x, len(Y)))
            new_y.append(Y)
        return np.concatenate(new_x), np.concatenate(new_y)

    def extract_spectra(self,
                        color=None,
                        use_exact_color=False,
                        image_range=None,
                        max_dominant_percentage=.2,
                        allow_grayscale=False,
                        color_space='rgb',
                        dominant_color_space='lab',
                        dominant_bins=255,
                        min_dominant_component=50,
                        extract_lines=True,
                        prune_frame_components=True,
                        frame_line_cutoffs=.5,
                        spectrum_direction='up',
                        x_range=(0, 1),
                        y_range=(0, 1),
                        preserve_x_range=True,
                        preserve_y_range=False,
                        use_entire_pixel_range=True,
                        return_color_code=True,
                        **opts
                        ):
        if isinstance(color, str):
            color = ColorPalette.parse_color_string(color)[:3]
            color_space = 'rgb'
        if color is None or not use_exact_color:
            opts = dev.OptionsSet(opts)
            dom_opts, opts = opts.split(None,
                                        {
                                            'bins',
                                            'merge_tolerances'
                                        }
                                        )
            doms = self.get_dominant_colors(dominant_bins,
                                            min_counts=min_dominant_component,
                                            color_space=dominant_color_space,
                                            **dom_opts)
            npixels = self.img.shape[-2] * self.img.shape[-1]
            max_npix = npixels * max_dominant_percentage
            if color is None:
                color = next((
                    d for d in doms
                    if (
                        doms[d] < max_npix
                        and (
                                allow_grayscale
                                or ColorPalette.color_convert(d, dominant_color_space, 'lch')[1] > .1
                        )
                    )
                ))
            else:
                c = ColorPalette.color_convert(color, color_space, dominant_color_space)
                color = next((
                    d for d in
                    sorted(doms,
                           key=lambda k: (np.linalg.norm(c - np.array(k))**2) * (doms[k] / npixels)
                           )
                ))
            color_space = dominant_color_space
        opts = dev.OptionsSet(opts)
        line_opts, pixel_opts = opts.split(None,
                                           {
                                               'max_pixel_distance',
                                               'min_line_cutoff',
                                               'smoothing',
                                               'line_split_cutoff',
                                               'allow_disjoint'
                                           }
                                           )
        y, x = self.find_pixels(color, image_range=image_range, color_space=color_space, **pixel_opts)
        if len(x) == 0 or len(y) == 0:
            return color, []

        if prune_frame_components:
            if prune_frame_components is True:
                prune_frame_components = [prune_frame_components, prune_frame_components]
            prune_x, prune_y = prune_frame_components
            x_bounds, y_bounds = self.identify_frame_boundaries(
                (x, y),
                identified_components=[prune_x, prune_y],
                min_line_cutoffs=frame_line_cutoffs
            )
            # raise Exception(x_bounds, y_bounds, (np.min(x), np.max(x)), (np.min(y), np.max(y)))
            mask = np.logical_and(
                np.logical_and(x_bounds[0] <= x,  x <= x_bounds[1]),
                np.logical_and(y_bounds[0] <= y,  y <= y_bounds[1])
            )

            x = x[mask]
            y = y[mask]

        y = self.img.shape[1] - y
        if extract_lines:
            lines = self.find_spectrum_lines((x, y), spectrum_direction=spectrum_direction, **line_opts)
        else:
            lines = [[x, y]]


        if extract_lines:
            if use_entire_pixel_range is True:
                use_entire_pixel_range = [True, False]
            if x_range is not None:
                if nput.is_int(x_range):
                    x_range = [0, x_range]

                if preserve_x_range:
                    if use_entire_pixel_range[0]:
                        x_span = [
                            np.min(x),
                            np.max(x)
                        ]
                    else:
                        x_mins = [
                            np.min(xx)
                            for xx, yy in lines
                            if len(xx) > 0
                        ]
                        if len(x_mins) == 0:
                            x_mins = [np.min(x)]

                        x_maxes = [
                            np.max(xx)
                            for xx, yy in lines
                            if len(xx) > 0
                        ]
                        if len(x_maxes) == 0:
                            x_maxes = [np.max(x)]

                        x_span = [
                            min(x_mins),
                            max(x_maxes)
                        ]
                    x_scale = (x_range[1] - x_range[0]) / (x_span[1] - x_span[0])
                    lines = [
                        [
                            (x - x_span[0]) * x_scale + x_range[0],
                            y
                        ]
                        for x,y in lines
                    ]
                else:
                    lines = [
                        [
                            (x - np.min(x)) * (x_range[1] - x_range[0]) / (np.max(x) - np.min(x)) + x_range[0],
                            y
                        ]
                        for x, y in lines
                    ]

            if y_range is not None:
                if nput.is_int(y_range):
                    y_range = [0, y_range]

                if preserve_y_range:
                    if use_entire_pixel_range[0]:
                        y_span = [
                            np.min(y),
                            np.max(y)
                        ]
                    else:
                        y_mins = [
                            np.min(yy)
                            for xx, yy in lines
                            if len(yy) > 0
                        ]
                        if len(y_mins) == 0:
                            y_mins = [np.min(y)]

                        y_maxes = [
                            np.max(yy)
                            for xx, yy in lines
                            if len(yy) > 0
                        ]
                        if len(y_maxes) == 0:
                            y_maxes = [np.max(y)]
                        y_span = [
                            min(y_mins),
                            max(y_maxes)
                        ]
                    y_scale = (y_range[1] - y_range[0]) / max([(y_span[1] - y_span[0]), 1])
                    lines = [
                        [
                            x,
                            (y - y_span[0]) * y_scale + y_range[0],
                        ]
                        for x,y in lines
                    ]
                else:
                    lines = [
                        [
                            x,
                            (y - np.min(y)) * (y_range[1] - y_range[0]) / (np.max(y) - np.min(y)) + y_range[0]
                        ]
                        for x, y in lines
                    ]

        if return_color_code:
            color = ColorPalette.rgb_code(
                ColorPalette.color_convert(
                    color, color_space, 'rgb'
                )
            )

        return color, lines






__all__ = [
    "Reaction"
]

import collections

import numpy as np
import McUtils.Numputils as nput
import McUtils.Devutils as dev
from ..Molecools import Molecule

from .ProfileGenerator import ProfileGenerator


class Reaction:
    def __init__(self,
                 reactants:list[Molecule],
                 products:list[Molecule],
                 align=True,
                 atom_mapping=None,
                 use_product_structure=True,
                 fragment_expansion_method=dev.default,
                 optimize=True,
                 energy_evaluator='rdkit',
                 profile_generator=None,
                 **fragment_initialization_opts
                 ):
        self.reactants = reactants
        self.products = products

        self.product_complex = self.make_reaction_complex(products,
                                                          fragment_expansion_method=fragment_expansion_method,
                                                          **fragment_initialization_opts
                                                          )
        if use_product_structure:
            self.reactant_complex = self.make_reaction_complex(reactants,
                                                               fragment_expansion_method=None)
        else:
            self.reactant_complex = self.make_reaction_complex(reactants,
                                                               fragment_expansion_method=fragment_expansion_method,
                                                               **fragment_initialization_opts
                                                               )

        if align:
            self.reactant_complex = self.product_complex.align_molecule(
                self.reactant_complex,
                align_structures=not use_product_structure,
                permute_atoms=False if use_product_structure else 'all'
            )

        if use_product_structure:
            if fragment_expansion_method is not None:
                method, opts = self.get_expansion_dispatch().resolve(fragment_expansion_method)
                target_coords = method(self.reactant_complex.fragment_indices,
                                       self.product_complex.coords,
                                       **dict(fragment_initialization_opts, **opts)
                                       )
            else:
                target_coords = self.product_complex.coords

            self.reactant_complex = self.reactant_complex.modify(coords=target_coords)

        if optimize:
            self.product_complex = self.product_complex.optimize(evaluator=energy_evaluator)
            self.reactant_complex = self.reactant_complex.optimize(evaluator=energy_evaluator)
            if align:
                self.reactant_complex = self.product_complex.align_molecule(
                    self.reactant_complex,
                    reindex_bonds=False,
                    align_structures=True,
                    permute_atoms=False if use_product_structure else 'all'
                )

        self.profile_generator = profile_generator

    def get_modified_bonds(self):
        return self.reactant_complex.edge_graph.graph_difference(
            self.product_complex.edge_graph
        )

    ReactionCore = collections.namedtuple("ReactionCore", ["indices", "formed", "broken"])
    def get_reaction_core(self):
        new, old = self.get_modified_bonds()
        return np.unique(np.concatenate([new, old], axis=0)), new, old

    @classmethod
    def expand_fragments_centroid(cls, inds, coords,
                                  min_distance=1,
                                  add_radius=True,
                                  expansion_factor=.1,
                                  max_iterations=5):
        coords = coords.copy()
        centroids = np.array([
            np.average(coords[i, :], axis=0)
            for i in inds
        ])
        rows, cols = np.triu_indices(len(centroids), k=1)
        initial_centroid_distances = np.linalg.norm(centroids[rows,] - centroids[cols,], axis=1)
        initial_centroids = centroids.copy()

        min_distance = [min_distance] * len(centroids)
        if add_radius:
            for i,x in enumerate(inds):
                rx, cx = np.triu_indices(len(x), k=1)
                rx = [x[j] for j in rx]
                cx = [x[j] for j in cx]
                radius = np.max(np.linalg.norm(coords[rx, :] - coords[cx, :], axis=1))
                min_distance[i] += radius

        for it_no in range(max_iterations):
            for k,(i,j) in enumerate(zip(rows, cols)):
                v = centroids[i] - centroids[j]
                v_norm = np.linalg.norm(v)
                if v_norm < 1e-4:
                    v = np.array([min_distance[k]/(1 + expansion_factor/2), 0, 0])
                di = v * (1 + expansion_factor/2)
                dj = -v * (1 + expansion_factor/2)
                centroids[i] += di
                centroids[j] += dj

            centroid_distances = np.linalg.norm(centroids[rows,] - centroids[cols,], axis=1)
            if all(
                    cd > md and cd >= (1+expansion_factor) * id
                    for md,cd,id in zip(min_distance, centroid_distances, initial_centroid_distances)
            ):
                break

        displacements = centroids - initial_centroids
        for idx, d in zip(inds, displacements):
            coords[idx,] += d[np.newaxis]

        return coords

    @classmethod
    def expand_fragments_nearest(cls, inds, coords, min_distance=5, max_iterations=15):
        coords = coords.copy()
        rows, cols = np.triu_indices(len(inds), k=1)
        rows = [inds[i] for i in rows]
        cols = [inds[j] for j in cols]

        for it_no in range(max_iterations):
            for (i, j) in zip(rows, cols):
                vecs = coords[i, np.newaxis, :] - coords[np.newaxis, j, :]
                dist = np.linalg.norm(vecs, axis=2)
                x,y = np.unravel_index(np.argmin(dist), (len(i), len(j)))
                md = dist[x, y]
                print(md)
                if md < min_distance:
                    v = vecs[x, y]
                    disp = v * (min_distance / md / 2)
                    coords[i,] += disp[np.newaxis]
                    coords[j,] -= disp[np.newaxis]

            nearest_distances = np.array([
                np.min(
                    np.linalg.norm(coords[i, np.newaxis, :] - coords[np.newaxis, j, :], axis=2)
                )
                for i,j in zip(rows, cols)
            ])

            if np.all(nearest_distances > min_distance):
                break

        return coords

    @classmethod
    def expand_fragments_attractive(cls,
                                    inds, coords,
                                    min_distance=5,
                                    max_displacement=.2,
                                    max_iterations=10,
                                    tol=1e-3
                                    ):
        coords = coords.copy()
        rows, cols = np.triu_indices(len(inds), k=1)
        rows = [inds[i] for i in rows]
        cols = [inds[j] for j in cols]

        for it_no in range(max_iterations):
            disp_norm = 0
            for (i, j) in zip(rows, cols):
                vecs = coords[i, np.newaxis, :] - coords[np.newaxis, j, :]
                dist = np.linalg.norm(vecs, axis=2)
                vecs = nput.vec_normalize(vecs, norms=dist)
                lj_param = min_distance/dist
                lj_grads = -(lj_param**13 - lj_param**7)[:, :, np.newaxis] * vecs
                disps = np.sign(lj_grads) * np.clip(np.abs(lj_grads), 0, max_displacement)

                d = np.sum(np.sum(disps, axis=0), axis=0)

                coords[i,] += d[np.newaxis]
                coords[j,] -= d[np.newaxis]

                disp_norm += np.max(np.abs(disps))

            if disp_norm < tol:
                break

        return coords

    @classmethod
    def get_expansion_methods(cls):
        return {
            'nearest': cls.expand_fragments_nearest,
            'attractive': cls.expand_fragments_attractive,
            'centroid': cls.expand_fragments_centroid
        }

    _expansion_dispatch = dev.uninitialized
    @classmethod
    def get_expansion_dispatch(cls):
        cls._expansion_dispatch = dev.handle_uninitialized(
            cls._expansion_dispatch,
            dev.OptionsMethodDispatch,
            args=(cls.get_expansion_methods,),
            kwargs=dict(default_method='attractive')
        )
        return cls._expansion_dispatch
    @property
    def expansion_methods(self):
        return self.get_expansion_dispatch()

    @classmethod
    def initialize_reactant_structures(cls,
                                       reactant_complex: Molecule,
                                       product_complex,
                                       fragment_expansion_method=None,
                                       **expansion_method_options
                                       ):
        if fragment_expansion_method is not None:
            fragment_indices = reactant_complex.fragment_indices
            method, opts = cls.get_expansion_dispatch().resolve(fragment_expansion_method)
            target_coords = method(fragment_indices, product_complex.coords, **dict(opts, **expansion_method_options))
        else:
            target_coords = product_complex.coords

        return reactant_complex.modify(coords=target_coords)

    @classmethod
    def concenate_molecules(cls, mols):
        atoms = sum([m.atoms for m in mols], ())
        masses = np.concatenate([m.masses for m in mols])
        coords = np.concatenate(
            [
                m.coords for m in mols
            ],
            axis=0
        )
        mol_shift = np.cumsum([0] + [len(m.atoms) for m in mols])
        bonds = sum(
            [
                [
                    [b[0] + shift, b[1] + shift] + list(b[2:])
                    for b in mol.bonds
                ]
                for mol, shift in zip(mols, mol_shift)
            ],
            []
        )

        return Molecule(
            atoms,
            coords,
            masses=masses,
            bonds=bonds
        )

    @classmethod
    def make_reaction_complex(cls,
                              mols,
                              coords=None,
                              fragment_expansion_method=None,
                              **expansion_method_options
                              ):
        reactant_complex = cls.concenate_molecules(mols)
        if coords is not None:
            reactant_complex = reactant_complex.modify(coords=coords)
        if fragment_expansion_method is not None:
            fragment_indices = reactant_complex.fragment_indices
            method, opts = cls.get_expansion_dispatch().resolve(fragment_expansion_method)
            target_coords = method(fragment_indices, reactant_complex.coords, **dict(opts, **expansion_method_options))
            reactant_complex = reactant_complex.modify(coords=target_coords)
        return reactant_complex

    @classmethod
    def from_smiles(cls, smiles, fragment_expansion_method=dev.default, align=True,
                    add_implicit_hydrogens=True,
                    **opts):
        reactants, products = [s.strip() for s in smiles.split(">>")]
        reactants = [
            Molecule.from_string(
                s.strip(), "smi", add_implicit_hydrogens=add_implicit_hydrogens
            )
            for s in reactants.split(".")
        ]

        products = [
            Molecule.from_string(
                s.strip(), "smi", add_implicit_hydrogens=add_implicit_hydrogens
            )
            for s in products.split(".")
        ]

        return cls(reactants, products,
                   fragment_expansion_method=fragment_expansion_method,
                   align=align,
                   **opts
                   )

    def get_profile_generator(self, profile_generator=None, internals=None, **profile_opts) -> ProfileGenerator:
        if profile_generator is None:
            profile_generator = self.profile_generator
        return ProfileGenerator.resolve(
            self.reactant_complex,
            self.product_complex,
            profile_generator=profile_generator,
            internals=internals,
            **profile_opts
        )
